@@ -3,15 +3,34 @@
 
 from collections import namedtuple
 import copy
+from dataclasses import dataclass
 from itertools import product
 import time
+from typing import Optional
 
-from .components import COMPONENT_SHAPES, Pipe, Component, Input, Output, Reactor, Recycler, DisabledOutput
-from .exceptions import RunSuccess, ReactionError
+import cursor
+from rich import print
+# TODO: Hoping to avoid default number-highlighting via Console(highlight=False).print but it seems to break ANSI resets
+
+from .components import Component, Input, Output, Reactor, Recycler, DisabledOutput
+from .waldo import InstructionType
+from .exceptions import ScoreError
 from .grid import *
 from .level import OVERWORLD_COLS, OVERWORLD_ROWS
 from .terrains import terrains, MAX_TERRAIN_INT
 
+@dataclass
+class DebugOptions:
+    """Debug options for running a solution.
+
+    reactor: The reactor to debug. If None, overworld is shown for production levels, or only reactor for researches
+    cycle: The cycle to start debugging at (0 for full debug)
+    speed: The speed to run the solution at
+    """
+    reactor: Optional[int]  # The reactor to debug. If None, overworld is shown for production levels, or only reactor
+                            # for research levels
+    cycle: int  # The cycle to start debugging at
+    speed: float  # The speed to run the solution at, in cycles / s
 
 class Score(namedtuple("Score", ('cycles', 'reactors', 'symbols'))):
     '''Immutable class representing a SpaceChem solution score.'''
@@ -23,21 +42,50 @@ class Score(namedtuple("Score", ('cycles', 'reactors', 'symbols'))):
 
     @classmethod
     def is_score_str(cls, s):
-        '''Return True if the given string is formatted like a Spacechem score, e.g. 45-1-14. 0-0-0 is included, but
-        note that it is used to denote incomplete solutions.
+        '''Return True if the given string is formatted like a Spacechem score, e.g. 45-1-14.
+        Incomplete score formats (0-0-0, and Incomplete-r-s from the old SCT) are included.
         '''
         parts = s.split('-')
-        return len(parts) == 3 and all(part.isdigit() for part in parts)  # Thinks 045-1-14 is a score but whatever
+        return (len(parts) == 3
+                and (parts[0].isdigit() or parts[0] == 'Incomplete') # Thinks 045-1-14 is a score but whatever
+                and all(part.isdigit() for part in parts[1:]))
 
     @classmethod
     def from_str(cls, s):
+        """Return a Score object, or None if the string represents an incomplete score."""
+        parts = s.split('-')
+        assert len(parts) == 3, "String is not a score"
+
+        # Return None politely on either modern SC format or old SaveChemTool format for incomplete solution
+        if s == '0-0-0' or (parts[0] == 'Incomplete' and all(part.isdigit() for part in parts[1:])):
+            return None
+
         return cls(*(int(x) for x in s.split('-')))
 
 
 class Solution:
     '''Class for constructing and running game entities from a given level object and solution code.'''
     __slots__ = ('level_name', 'author', 'expected_score', 'name',
-                 'level', 'components')
+                 'level', 'components', 'cycle')
+
+    DEFAULT_MAX_CYCLES = 1_000_000  # Default timeout for solutions whose expected cycle count is unknown
+
+    # Convenience properties
+    @property
+    def inputs(self):
+        return (component for component in self.components if isinstance(component, Input))
+
+    @property
+    def reactors(self):
+        return (component for component in self.components if isinstance(component, Reactor))
+
+    @property
+    def outputs(self):
+        return (component for component in self.components if isinstance(component, Output))
+
+    @property
+    def description(self):
+        return self.describe(self.level.name, self.author, self.expected_score, self.name)
 
     @classmethod
     def parse_metadata(cls, s):
@@ -51,9 +99,10 @@ class Solution:
         it is possible for some fields to get mis-parsed. This is still a little better than SC which gets
         completely messed up by any commas in the level name.
         '''
-        s = s.strip().split('\n', maxsplit=1)[0].strip()
-        assert s.startswith('SOLUTION:'), "Given string is not a SpaceChem solution"
-        fields = s[len('SOLUTION:'):].split(',')
+        first_line = s.replace('\r\n', '\n').strip('\n').split('\n', maxsplit=1)[0]  # Get first non-empty line
+
+        assert first_line.startswith('SOLUTION:'), "Given string is not a SpaceChem solution"
+        fields = first_line[len('SOLUTION:'):].split(',')
         assert len(fields) >= 3, "Missing fields in solution metadata"
 
         # Starting from the third CSV value, look for a score-like string and assume it is the expected score
@@ -61,19 +110,24 @@ class Solution:
                                None)
         assert score_field_idx is not None, "Solution metadata missing expected score"
 
+        # Assume any excess commas preceding the score are part of the level name
         level_name = ','.join(fields[:score_field_idx - 1])
         author_name = fields[score_field_idx - 1]
-        # The game stores unsolved solutions as '0-0-0'
 
-        expected_score = Score.from_str(fields[score_field_idx]) if fields[score_field_idx] != '0-0-0' else None
+        expected_score = Score.from_str(fields[score_field_idx])  # Note that this returns None for an incomplete score
         soln_name = ','.join(fields[score_field_idx + 1:]) if len(fields) > score_field_idx + 1 else None
+        # Game single-quotes solution names if they contain a comma, strip this
+        if soln_name is not None and ',' in soln_name and soln_name[0] == soln_name[-1] == "'":
+            soln_name = soln_name[1:-1]
 
         return level_name, author_name, expected_score, soln_name
 
     @classmethod
     def describe(cls, level_name, author, expected_score, soln_name):
         """Return a pretty-printed string describing the given solution metadata."""
-        soln_descr = f"[{level_name}] {expected_score}"
+        soln_descr = f"[{level_name}]"
+        if expected_score is not None:
+            soln_descr += f' {expected_score}'
         if soln_name is not None:
             soln_descr += f' "{soln_name}"'
         soln_descr += f" by {author}"
@@ -81,20 +135,27 @@ class Solution:
         return soln_descr
 
     @classmethod
-    def split_solutions(cls, soln_str):
-        """Given a string potentially containing multiple solutions, return an iterator of them."""
-        soln_str = soln_str.strip()
-        assert soln_str.startswith('SOLUTION:'), "Given text is not a SpaceChem solution"
+    def split_solutions(cls, solns_str):
+        """Given a string potentially containing multiple solutions, return an iterator of them.
+        Returns an empty iterator if the given string has no non-empty lines.
+        """
+        solns_str = '\n'.join(s for s in solns_str.replace('\r\n', '\n').split('\n') if s)  # Remove empty lines
+
+        # Ensure first non-empty line is a solution string
+        if solns_str and not solns_str.startswith('SOLUTION:'):
+            raise ValueError("Invalid solution string: expected SOLUTION: on line 1")
 
         # Split with newline prefix in case some fucker names themselves SOLUTION:
-        return (f'SOLUTION:{s}'.strip() for s in ('\n' + soln_str).split('\nSOLUTION:')[1:])
+        return (f'SOLUTION:{s}' for s in ('\n' + solns_str).split('\nSOLUTION:')[1:])
 
+    # TODO: Solution constructor should probably accept either level or level_code for convenience
     def __init__(self, level, soln_export_str=None):
         self.level = level
-        self.level_name = level['name']  # Note that the level name is vestigial; a solution can run in any compatible level
+        self.level_name = level['name']
         self.name = None
         self.author = 'Unknown'
         self.expected_score = None
+        self.cycle = 0
 
         # Set up the level terrain so we can look up input/output positions and any blocked terrain
         if level['type'].startswith('production'):
@@ -159,7 +220,8 @@ class Solution:
             else:
                 # CE production levels expect input types/positions to be manually specified in the input dict
                 for input_dict in self.level[input_zone_type]:
-                    if not input_dict['molecules']:
+                    # Skip empty random input zones
+                    if 'molecules' in input_dict and not input_dict['molecules']:
                         continue
 
                     new_component = Input(input_dict=input_dict, is_research=self.level['type'].startswith('research'))
@@ -229,18 +291,26 @@ class Solution:
 
         # Add solution-defined components and update preset level components (e.g. inputs' pipes, preset reactor contents)
         if soln_export_str is not None:
-            # Parse solution metadata from the first line
-            soln_metadata_str, components_str = soln_export_str.strip().split('\n', maxsplit=1)
+            # Get the first non-empty line
+            soln_metadata_str, *split_remainder = soln_export_str.replace('\r\n', '\n').strip('\n').split('\n', maxsplit=1)
+            components_str = '' if not split_remainder else split_remainder[0].strip('\n')
             self.level_name, self.author, self.expected_score, self.name = self.parse_metadata(soln_metadata_str)
 
-            for component_str in ('COMPONENT:' + s for s in components_str.split('COMPONENT:') if s):
-                component_metadata = component_str.split('\n', maxsplit=1)[0]
-                fields = component_metadata.split(',')
-                component_type = fields[0].strip('COMPONENT:').strip("'")
-                component_posn = Position(int(fields[1]), int(fields[2]))
+            if components_str:
+                assert components_str.startswith('COMPONENT:'), "Unexpected data on line 1"
+
+            soln_defined_component_posns = set()  # Used to check that the solution doesn't doubly-define a component
+
+            for component_str in ('COMPONENT:' + s for s in ('\n' + components_str).split('\nCOMPONENT:')[1:]):
+                component_type, component_posn = Component.parse_metadata(component_str)
+
+                # Ensure this component hasn't already been created/updated by the solution
+                if component_posn in soln_defined_component_posns:
+                    raise ValueError(f"Solution defines component at {component_posn} twice")
+                soln_defined_component_posns.add(component_posn)
 
                 # Create a raw instance of the component if it doesn't exist yet
-                if not component_posn in posn_to_component:
+                if component_posn not in posn_to_component:
                     # Ensure this is a legal component type for the level (e.g. drag-starter-reactor -> has-starter)
                     reactor_type_flag = f'has-{component_type.split("-")[1]}'
                     assert ((reactor_type_flag in self.level and self.level[reactor_type_flag])
@@ -282,7 +352,14 @@ class Solution:
                 update_pipes = (not self.level['type'].startswith('research')
                                 and ('mutable-pipes' not in self.level
                                      or self.level['mutable-pipes']))
-                posn_to_component[component_posn].update_from_export_str(component_str, update_pipes=update_pipes)
+                component = posn_to_component[component_posn]
+                try:
+                    component.update_from_export_str(component_str, update_pipes=update_pipes)
+                except Exception as e:
+                    if not self.level['type'].startswith('research'):
+                        raise type(e)(f"{component.type} at {component.posn}: {e}")
+
+                    raise e
 
                 # TODO: Ensure the updated component pipes are within the overworld bounds
 
@@ -353,12 +430,12 @@ class Solution:
                         pipe_posns_to_dirns[real_posn] = set()
 
                     # Pipe is not vertical (blocks the horizontal direction)
-                    if not (prev.col == cur.col == next_.col):
+                    if not prev.col == cur.col == next_.col:
                         assert RIGHT not in pipe_posns_to_dirns[real_posn], f"Illegal pipe overlap at {real_posn}"
                         pipe_posns_to_dirns[real_posn].add(RIGHT)
 
                     # Pipe is not horizontal (blocks the vertical direction)
-                    if not (prev.row == cur.row == next_.row):
+                    if not prev.row == cur.row == next_.row:
                         assert UP not in pipe_posns_to_dirns[real_posn], f"Illegal pipe overlap at {real_posn}"
                         pipe_posns_to_dirns[real_posn].add(UP)
 
@@ -409,7 +486,11 @@ class Solution:
         # Solution metadata
         export_str = f"SOLUTION:{self.level['name']},{self.author},{self.expected_score}"
         if self.name is not None:
-            export_str += f',{self.name}'
+            # SC expects single-quotes surrounding solution names that contain a comma
+            if ',' in self.name:
+                export_str += f",'{self.name}'"
+            else:
+                export_str += f",{self.name}"
 
         # Components
         # Exclude inputs whose pipes are length 1 (unmodified), outputs, and the recycler.
@@ -417,16 +498,15 @@ class Solution:
         # TODO: This doesn't cover inputs with preset pipes > 1 long - which also shouldn't be included
         #       Really it's probably just every unmodified preset component
         for component in self.components:
-            if not (isinstance(component, Output)
+            if not (isinstance(component, (Output, Recycler))
                     or (isinstance(component, Input)
-                        and len(component.out_pipe) == 1)
-                    or isinstance(component, Recycler)):
+                        and len(component.out_pipe) == 1)):
                 export_str += '\n' + component.export_str()
 
         return export_str
 
     def __str__(self):
-        ''''Return a string representing the overworld of this solution including all components and pipes.'''
+        '''Return a string representing the overworld of this solution including all components/pipes, and the cycle.'''
         # 1 character per tile
         grid = [[' ' for _ in range(OVERWORLD_COLS)] for _ in range(OVERWORLD_ROWS)]
 
@@ -461,16 +541,29 @@ class Solution:
         result = f"_{OVERWORLD_COLS * '_'}_\n"
         for row in grid:
             result += f"|{''.join(row)}|\n"
-        result += f"‾{OVERWORLD_COLS * '‾'}‾\n"
+        result += f"‾{OVERWORLD_COLS * '‾'}‾"
+
+        result += f'\nCycle: {self.cycle}'
 
         return result
 
-    def debug_print(self, cycle, duration=0.5):
+    def debug_print(self, duration=0.5, reactor_idx=None, flash_features=True):
+        '''Print the currently running solution state then clear it from the terminal.
+        Args:
+            duration: Seconds before clearing the printout from the screen. Default 0.5.
+            reactor_idx: If specified, only print the contents of the reactor with specified index.
+            flash_features: Whether to flash activated features, inputs, and outputs.
+        '''
+        if reactor_idx is None:
+            output = str(self)
+        else:
+            output = list(self.reactors)[reactor_idx].__str__(flash_features=flash_features)
+            output += f'\nCycle: {self.cycle}'
+
         # Print the current state
-        output = str(self)
-        output += f'\nCycle: {cycle}'
         print(output)  # Could use end='' but that makes keyboard interrupt output ugly
 
+        # TODO: this sleep is additive with the time schem takes to run each cycle so debug is always too slow
         time.sleep(duration)
 
         # Use the ANSI escape code for moving to the start of the previous line to reset the terminal cursor
@@ -486,7 +579,21 @@ class Solution:
             assert sum(1 if isinstance(component, Reactor) else 0
                        for component in self.components) <= self.level['max-reactors'], "Reactor limit exceeded"
 
-    def run(self, max_cycles=1000000, debug=False):
+    def cycle_movement(self):
+        """Move contents of all components one cycle's-worth forward."""
+        for component in self.components:
+            try:
+                component.move_contents(self.cycle)
+            except Exception as e:
+                # Mention the originating reactor in errors when possible
+                if isinstance(component, Reactor):
+                    for i, reactor in enumerate(self.reactors):
+                        if component is reactor:
+                            # Mention the reactor index via a chained exception of the same type
+                            raise type(e)(f"Reactor {i}: {e}") from e
+                raise e
+
+    def run(self, max_cycles=None, debug=False):
         '''Run this solution, returning a score tuple or else raising an exception if the level was not solved.'''
         # TODO: Running the solution should not meaningfully modify it. Namely, need to reset reactor.molecules
         #       and waldo.position/waldo.direction before each run, or otherwise prevent these from persisting.
@@ -495,78 +602,65 @@ class Solution:
         # TODO: Should re-check for component/pipe collisions every time we run it; run() should be a source of truth,
         #       and shouldn't be confoundable by modifying the solution after the ctor validations
 
+        # Set the maximum runtime if unspecified to ensure a broken solution can't infinite loop forever
+        if max_cycles is None:
+            if self.expected_score is not None:
+                max_cycles = 2 * self.expected_score.cycles
+            else:
+                max_cycles = self.DEFAULT_MAX_CYCLES
+
         # Default debug view to the reactor's interior in research levels
         if debug and self.level['type'].startswith('research'):
             debug.reactor = 0
 
         self.validate_components()
+        reactors = list(self.reactors)
+        outputs = list(self.outputs)
 
-        reactors = [component for component in self.components if isinstance(component, Reactor)]
-
-        # Set the maximum runtime to ensure a broken solution can't infinite loop forever
-        if self.expected_score is not None:
-            max_cycles = 2 * self.expected_score.cycles
+        if any(waldo.instr_map[waldo.position][1].type == InstructionType.PAUSE
+               for reactor in reactors
+               for waldo in reactor.waldos
+               if waldo.position in waldo.instr_map):
+            self.cycle_movement()
 
         # Run the level
         symbols = sum(sum(len(waldo) for waldo in component.waldos)
                       for component in self.components
                       if hasattr(component, 'waldos'))  # hacky but saves on using a counter or reactor list in the ctor
-        cycle = 0
-        completed_outputs = 0
-        num_outputs = len([c for c in self.components if isinstance(c, Output) and not isinstance(c, DisabledOutput)])
 
         try:
-            while cycle < max_cycles:
-                # Move molecules/waldos
-                for component in self.components:
-                    try:
-                        component.move_contents()
-                    except Exception as e:
-                        # Mention the originating reactor in errors when possible
-                        if isinstance(component, Reactor):
-                            for i, reactor in enumerate(reactors):
-                                if component is reactor:
-                                    # Mention the reactor index via a chained exception of the same type
-                                    raise type(e)(f"Reactor {i}: {e}") from e
-                        raise e
+            # In debug mode the cursor can annoyingly flicker into the middle of the printed output; hide it
+            if debug:
+                cursor.hide()
 
-                if debug and cycle >= debug.cycle:
-                    if debug.reactor is None:
-                        self.debug_print(cycle, duration=0.0625 / debug.speed)
-                    else:
-                        reactors[debug.reactor].debug_print(cycle, duration=0.0625 / debug.speed)
+            while self.cycle < max_cycles + 1:
+                self.cycle += 1
 
                 # Execute instant actions (entity inputs/outputs, waldo instructions)
                 for component in self.components:
-                    if component.do_instant_actions(cycle):
-                        # Outputs return true the first time they reach their target count; count these occurrences and
-                        # raise success when they've all completed
-                        completed_outputs += 1
-                        if completed_outputs == num_outputs:
-                            raise RunSuccess()
+                    if component.do_instant_actions(self.cycle):
+                        # Outputs return true the first time they reach their target count; whenever one does, check
+                        # all the others for completion (we can't just count completions since 0/0 outputs should
+                        # win even if never triggered). Not checking until an output completes also matches the expected
+                        # behaviour of puzzles where all outputs are 0/0
+                        if all(output.current_count >= output.target_count for output in outputs):
+                            # TODO: Update solution expected score? That would match the game's behavior, but makes the validator
+                            #       potentially misleading. Maybe run() and validate() should be the same thing.
+                            # -1 looks weird but seems provably right based on output vs pause comparison
+                            return Score(self.cycle - 1, len(reactors), symbols)
 
-                if debug and cycle >= debug.cycle:
-                    if debug.reactor is None:
-                        self.debug_print(cycle, duration=0.0625 / debug.speed)
-                    else:
-                        reactors[debug.reactor].debug_print(cycle, duration=0.0625 / debug.speed)
+                if debug and self.cycle >= debug.cycle:
+                    self.debug_print(duration=0.7 / debug.speed, reactor_idx=debug.reactor)
 
-                cycle += 1
+                self.cycle_movement()
+
+                if debug and self.cycle >= debug.cycle:
+                    self.debug_print(duration=0.3 / debug.speed, reactor_idx=debug.reactor, flash_features=False)
 
             raise TimeoutError(f"Solution exceeded {max_cycles} cycles, probably infinite looping?")
-        except RunSuccess:
-            # TODO: Update solution expected score? That would match the game's behavior, but makes the validator
-            #       potentially misleading. Maybe run() and validate() should be the same thing.
-            # TODO: The cycle + 1 here is a hack since we're inconsistent with SC's displayed count. Given that
-            #       waldos move one tile before inputs populate their pipe (and the cycle is already displayed as 1
-            #       while they do that first move), I think the only way to match SC exactly is to do instant actions
-            #       before move actions, but to have inputs trigger when `(cycle - 1) % rate == 0` instead of
-            #       `cycle % rate == 0`, so that they don't input on cycle 0 (before the waldos have moved).
-            #       For now putting the +1 here looks less awkward...
-            return Score(cycle + 1, len(reactors), symbols)
         except Exception as e:
-            # Mention the cycle number on error via a chained exception of the same type
-            raise type(e)(f"Cycle {cycle}: {e}") from e
+            # Mention the solution description and cycle number on error via a chained exception of the same type
+            raise type(e)(f"{self.description}: Cycle {self.cycle}: {e}") from e
         finally:
             # Persist the last debug printout
             if debug:
@@ -574,16 +668,30 @@ class Solution:
                     print(str(self))
                 else:
                     print(str(reactors[debug.reactor]))
+                    print(f'Cycle: {self.cycle}')
 
-    def validate(self, verbose=True, debug=False):
-        '''Run this solution and assert that the resulting score matches the expected score that was included in its
-        solution code.
-        '''
-        if self.level_name != self.level.name:
+                # Restore the cursor
+                cursor.show()
+
+    def validate(self, max_cycles=None, verbose=False, debug=False):
+        '''Run this solution and assert that the score matches the expected score from its metadata.'''
+        if self.expected_score is None:
+            raise ValueError("validate() requires a valid expected score in the first solution line (currently 0-0-0);"
+                             + " please update it or use run() instead.")
+
+        if verbose and self.level_name != self.level.name:
             print(f"Warning: Validating solution against level {repr(self.level.name)} that was originally"
                   + f" constructed for level {repr(self.level_name)}.")
 
-        score = self.run(debug=debug)
-        assert score == self.expected_score, f"Expected score {self.expected_score} but got {score}"
+        if max_cycles is None:
+            max_cycles = 2 * self.expected_score.cycles
+        elif self.expected_score.cycles > max_cycles:
+            raise ValueError(f"{self.description}: Cannot validate; expected cycles > max cycles ({max_cycles})")
+
+        score = self.run(max_cycles=max_cycles, debug=debug)
+
+        if score != self.expected_score:
+            raise ScoreError(f"{self.description}: Expected score {self.expected_score} but got {score}")
+
         if verbose:
-            print(f"Validated {describe(self)}")
+            print(f"Validated {self.description}")
